/*
 *  CoroX.c
 *  Coroutines for Mac OS X
 *
 *  Created by Jens Alfke on 4/29/08.
 *  Adapted from Steve Dekorte's libCoroutine:
 *  <http://www.dekorte.com/projects/opensource/libCoroutine/>
 *  by putting it on a piece of wood and banging a few nails through it.
 *  No, actually I removed all the stuff for cross-platform support, leaving only the simple
 *  code that works on Mac OS X 10.5, and then cleaned things up a bit.
 *
 *  Copyright 2008 Jens Alfke. All rights reserved.
 *  Copyright (c) 2002, 2003 Steve Dekorte. All rights reserved.
 *  License is at the bottom of this file.
 *
 */

#include "CoroX.h"
#include <ucontext.h>
#include <stddef.h>
#include <stdio.h>
#include <assert.h>


#define CORO_DEFAULT_STACK_SIZE (65536*4)
#define CORO_STACK_SIZE_MIN     8192


struct Coro
{        
    size_t stackSize;
    void *stack;
    ucontext_t env;
    // The following field works around bug(?) in sys/_structs.h. This field should be part of ucontext_t:
    _STRUCT_MCONTEXT	env_registers; 
    unsigned char isMain;
};


typedef struct CallbackBlock
{
    void *context;
    CoroStartCallback *func;
} CallbackBlock;


Coro *Coro_new(void)
{
    Coro *self = (Coro *)calloc(1, sizeof(Coro));
    self->stackSize = CORO_DEFAULT_STACK_SIZE;
    self->stack = NULL;
    return self;
}

void Coro_free(Coro *self)
{
    if (self->stack)
        free(self->stack);
    free(self);
}


void *Coro_stack(Coro *self)
{
    return self->stack;
}

size_t Coro_stackSize(Coro *self)
{
    return self->stackSize;
}

void Coro_setStackSize_(Coro *self, size_t sizeInBytes)
{
    self->stackSize = sizeInBytes;
}

uint8_t *Coro_CurrentStackPointer(void) __attribute__ ((noinline));

uint8_t *Coro_CurrentStackPointer(void)
{
    uint8_t a;
    uint8_t *b = &a;
    return b;
}

size_t Coro_bytesLeftOnStack(Coro *self)
{
    uint8_t dummy;
    ptrdiff_t p1 = (ptrdiff_t)(&dummy);
    ptrdiff_t p2 = (ptrdiff_t)Coro_CurrentStackPointer();
    int stackMovesUp = p2 > p1;
    ptrdiff_t start = ((ptrdiff_t)self->stack);
    ptrdiff_t end = start + self->stackSize;
    
    if (stackMovesUp)
    {
        return end - p1;
    }
    else
    {
        return p1 - start;
    }
}

int Coro_stackSpaceAlmostGone(Coro *self)
{
    return Coro_bytesLeftOnStack(self) < CORO_STACK_SIZE_MIN;
}


void Coro_initializeMainCoro(Coro *self)
{
    self->isMain = 1;
}


static void Coro_StartWithArg(CallbackBlock *block)
{
    (block->func)(block->context);
    fprintf(stderr,"CoroX error: returned from coro start function\n");
    abort();
}

typedef void (*makecontext_func)(void);

static void Coro_setup(Coro *self, void *arg)
{
    ucontext_t *ucp = (ucontext_t *) &self->env;
    getcontext(ucp);
    
    ucp->uc_stack.ss_sp = Coro_stack(self);
    ucp->uc_stack.ss_size = Coro_stackSize(self);
    ucp->uc_stack.ss_flags = 0;
    ucp->uc_link = NULL;
    
    makecontext(ucp, (makecontext_func)Coro_StartWithArg, 1, arg);
}


void Coro_startCoro_(Coro *self, Coro *other, void *context, CoroStartCallback *callback)
{
    assert(other->stack==NULL);
    other->stack = calloc(1, other->stackSize + 16);

    CallbackBlock callbackBlock = {context,callback};
    Coro_setup(other, &callbackBlock);
    
    Coro_switchTo_(self, other);
}

void Coro_switchTo_(Coro *self, Coro *next)
{
    swapcontext(&self->env, &next->env);
}




/*
 (This is a BSD License)
 
 Copyright (c) 2002, 2003 Steve Dekorte
 Copyright (c) 2008 Jens Alfke
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without modification, are
 permitted provided that the following conditions are met:
 
 • Redistributions of source code must retain the above copyright notice, this list of
 conditions and the following disclaimer.
 • Redistributions in binary form must reproduce the above copyright notice, this list
 of conditions and the following disclaimer in the documentation and/or other materials
 provided with the distribution.
 • Neither the name of the author nor the names of other contributors may be used to
 endorse or promote products derived from this software without specific prior written
 permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
