//
//  MYCoroutine.h
//  Coroutines for Mac OS X
//
//  Created by Jens Alfke on 4/29/08.
//  Copyright 2008 Jens Alfke. All rights reserved.
//

#import <Foundation/Foundation.h>


/** Objective-C coroutine class.
    See: http://en.wikipedia.org/wiki/Coroutine 
*/
@interface MYCoroutine : NSObject
{
    @private
    struct Coro *_coro;
    NSString *_name;
}

/** The "main" coroutine: the one corresponding to the execution context before the
    first coroutine is started. */
+ (MYCoroutine*) mainCoroutine;

/** The currently active coroutine. */
+ (MYCoroutine*) currentCoroutine;

/** Creates a new coroutine and starts it, performing the given invocation on it.
    That method must not return! */
+ (MYCoroutine*) startWithInvocation: (NSInvocation*)invocation;

/** Creates but does not start a coroutine. */
- (id) init;

/** Starts a new coroutine, and performs the given invocation on it.
    The current coroutine will block (i.e. this call won't return)
    until some other coroutine tells it to resume. */
- (void) startWithInvocation: (NSInvocation*)invocation;

/** Starts a new coroutine, invoking its -main method as its body.
    Since the default implementation of -main is empty, this only makes sense to call
    on an instance of a subclass of MYCoroutine that's overridden -main.
    The current coroutine will block (i.e. this call won't return)
    until some other coroutine tells it to resume. */
- (void) start;

/** The "main" method that will be called if the coroutine is started with -start.
    The default implementation is empty, so a subclass using -start must override this. */
- (void) main;

/** Returns control to an already-started (but blocked) coroutine.
    The most recent -resume call made from within that coroutine will return.
    The current coroutine will block (i.e. this call won't return)
    until some other coroutine tells it to resume. */
- (void) resume;


/** The coroutine's name. You can put anything you want here, as a debugging aid. */
@property (copy) NSString* name;

/** Returns YES if this is the currently executing coroutine. */
@property (readonly) BOOL isCurrent;

/** The stack size of the coroutine. You can only change this before calling -start! */
@property size_t stackSize;

/** The number of bytes of stack space left on this coroutine. */
@property (readonly) size_t bytesLeftOnStack;

/** Returns YES if this coroutine is almost out of stack space (less than 8k left) */
@property (readonly) BOOL stackSpaceAlmostGone;

@end
